/*
Copyright (C) 2001-2002 Charles Hollemeersch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

extern "C"
{
#include "../quakedef.h"
}

#define CLASSIC_MAX_QPATH 64
#define CLASSIC_MAXALIASTRIS 2048


#include <pspgu.h>
#include <pspgum.h>
//#include "video_hardware_md3.h"

int loadtextureimage (char* filename, int matchwidth, int matchheight, qboolean complain, int filter);

#define MULTIMODEL_LINK_STANDARD	0 //Just replace entity->model with the multimodel and draw.
#define MULTIMODEL_LINK_TAG			1 //Replace entity->model with the multimodel, translate according to the tag, and draw.
#define MULTIMODEL_LINK_INVTAG		2 //Like LINK_TAG, but the tag data is retrieved from the own model, not the linkedmodel.

#define MULTIMODEL_FL_ORIGIN		1
#define MULTIMODEL_FL_ANGLES		2
#define MULTIMODEL_FL_NODRAW		4
#define MULTIMODEL_FL_INVERT		8 //Invert offset & rotation
#define MULTIMODEL_FL_MODELLINKED	128 //The linkedmodel is a model-specific one.
typedef struct multimodel_s {
	char identifier[64];
	struct model_s *model;				//The actual model
	struct multimodel_s *linkedmodel;	//NULL if no special linking is neccesary.
	struct multimodel_s *nextmodel;

	byte linktype;				//0 if no special linking is neccesary.
	short flags;				//Flags to determine what data is stored.
	char tagname[64];		//Q3 tags are 64 chars, HL are only 32.
	vec3_t origin;
	vec3_t angles;
} multimodel_t;



#define TEMPHACK 1
#ifdef TEMPHACK
extern cvar_t temp1;
#endif

typedef float vec_t;
typedef vec_t vec2_t[2];//JTR: added additional vectors
void R_RotateForEntity (entity_t *e, int shadow);
void GL_Bind (int texnum);

#include <list>

using namespace std;

extern list<int> mapTextureNameList;


typedef struct
{
	char    name[CLASSIC_MAX_QPATH];
	int             filepos, filelen;
} packfile_t;

typedef struct pack_s
{
	char    filename[MAX_OSPATH];
	int             handle;
	int             numfiles;
	packfile_t      *files;
} pack_t;


typedef struct searchpath_s
{
	char    filename[MAX_OSPATH];
	pack_t  *pack;          // only one of filename / pack will be used
	struct searchpath_s *next;
} searchpath_t;




typedef struct searchpath_s searchpath_t;
extern searchpath_t* com_searchpaths;

/*
====================================
COM_MultipleSearch
takes a searchstring with wildcards (*) and searches for all files maching it.
returns an array of strings that need to be cleared.
====================================
*/
#include <pspiofilemgr.h>

int COM_MultipleSearch (char *searchstring, char **found, int max, qboolean fullpath)
{
	int				i;
	searchpath_t	*search = com_searchpaths;
	char			basestring[CLASSIC_MAX_QPATH];
	int				baselen, numfound = 0;

	baselen = (int)COM_SkipPath(searchstring) - (int)searchstring;
	Q_strncpy(basestring, searchstring, baselen);

	for ( ; search; search = search->next)
	{
		if (search->pack)
		{
		// look through all the pak file elements
			pack_t *pak = search->pack;
			if(fullpath)
				continue;
			for (i=0 ; i<pak->numfiles && numfound < max ; i++)
				if (!Q_strcmp (basestring, pak->files[i].name))
				{       // found a matching mapname.. now to match wildcards.
					char *substr = pak->files[i].name;
					char *wildcardstring = COM_SkipPath(searchstring);
					int n = 0; //postition in wildcardstring

					while(1)
					{
						char tmp[CLASSIC_MAX_QPATH];
						int s = 0;

						switch(wildcardstring[n])
						{
						case '\0':
							if(substr[0])
								goto next;
							else
							{//Found it!
								found[numfound] = static_cast<char*>(Z_Malloc(strlen(pak->files[i].name)*sizeof(char)));
								Q_strcpy(found[numfound],pak->files[i].name);
								numfound++;
							}

						case '*':
							while(wildcardstring[n] != '\0' && wildcardstring[n] != '*' && wildcardstring[n] != '?')
								tmp[s++] = substr[n++];
							tmp[s] = '\0';
							substr = strstr(tmp, substr);
								if(!substr)
									goto next;
						default:
							while(wildcardstring[n] != '\0' && wildcardstring[n] != '*' && wildcardstring[n] != '?')
								tmp[s++] = substr[n++];
							tmp[s] = '\0';
							if(Q_strcmp(tmp, substr))
								goto next;
							substr += s-1;
						}
						n++;
						substr++;
					}
					found[numfound] = static_cast<char*>(Z_Malloc(strlen(pak->files[i].name)*sizeof(char)));
					Q_strcpy(found[numfound], pak->files[i].name);
					numfound++;
next:;
				}
		}
		else
		{               
			Con_Printf("Sorry COM_MultipleSearch not worked\n");
/*
			char			filestring[CLASSIC_MAX_QPATH];
			char				dirstring [CLASSIC_MAX_QPATH];
			struct _finddata_t	fileinfo;
			int					handle;

			sprintf (dirstring, "%s/%s", search->filename, searchstring);
			handle = _findfirst (dirstring, &fileinfo);
			if (handle != -1)
			do
			{
				if (fileinfo.name[0] == '.')
					continue;

				if(fullpath)
					sprintf(filestring,"%s/%s%s", search->filename, basestring, fileinfo.name);
				else
					sprintf(filestring,"%s%s", basestring, fileinfo.name);
				found[numfound] = Z_Malloc(strlen(filestring)*sizeof(char));
				Q_strcpy(found[numfound],filestring);
				numfound++;
			}
			while (_findnext( handle, &fileinfo ) != -1 && numfound < max);
			_findclose (handle);
*/
		}
	}
	return numfound;
}



void R_DrawQ3Model(entity_t *e)
{
	md3header_t *model = static_cast<md3header_t*>(Mod_Extradata(e->model));
	int i, j;
	unsigned int *elem;
	md3mesh_t *mesh;
	float l;
	float	shadelight, ambientlight;

	ambientlight = shadelight = 200;//R_LightPoint (currententity->origin);

	// allways give the gun some light
	if (e == &cl.viewent && ambientlight < 24)
		ambientlight = shadelight = 24;

	for (i=0 ; i<MAX_DLIGHTS ; i++)
	{
		if (cl_dlights[i].die >= cl.time)
		{
			vec3_t dist;
			float add;
			VectorSubtract (currententity->origin,
							cl_dlights[i].origin,
							dist);
			add = cl_dlights[i].radius - Length(dist);

			if (add > 0) {
				ambientlight += add;
				//ZOID models should be affected by dlights as well
				shadelight += add;
			}
		}
	}

	// clamp lighting so it doesn't overbright as much
	if (ambientlight > 128)
		ambientlight = 128;
	if (ambientlight + shadelight > 192)
		shadelight = 192 - ambientlight;

	// ZOID: never allow players to go totally black
	i = currententity - cl_entities;
	if (i >= 1 && i<=cl.maxclients /* && !strcmp (currententity->model->name, "progs/player.mdl") */)
		if (ambientlight < 8)
			ambientlight = shadelight = 8;

	shadelight = shadelight / 200.0;
	l = shadelight;

	//glColor3f (l, l, l);

	sceGumPushMatrix();

	R_RotateForEntity(e);

	sceGumUpdateMatrix();

	mesh = (md3mesh_t *)((byte *)model + model->mesh_offs);

	for (i = 0; i < model->nummeshes; i++)
	{
		int numelems, frameoffset;
		frameoffset = e->frame;

		GL_Bind(((int *)((byte *)model + mesh->skin_offs))[e->skinnum%mesh->numskins]);

		
		elem = (unsigned int *)((byte *)mesh + mesh->elem_offs);
		numelems = mesh->numtris*3;
		if(e->frame > model->numboneframes)
			frameoffset = model->numboneframes-1;
		frameoffset *= mesh->numverts;

		// Allocate the vertices.
		struct vertex
		{
			float u, v;
			unsigned int color;
			float x, y, z;
		};

		vertex* const out = static_cast<vertex*>(sceGuGetMemory(sizeof(vertex) * numelems));

		for (j = 0; j < numelems; j++)
		{
			out[j].u = ((vec2_t *)((byte *)mesh + mesh->tc_offs))[*elem][0];
            out[j].v = ((vec2_t *)((byte *)mesh + mesh->tc_offs))[*elem][1];

			out[j].x =((vec3_t *)((byte *)model + mesh->vert_offs))[*elem + frameoffset][0];
            out[j].y =((vec3_t *)((byte *)model + mesh->vert_offs))[*elem + frameoffset][1];
            out[j].z =((vec3_t *)((byte *)model + mesh->vert_offs))[*elem + frameoffset][2];

			out[j].color = GU_COLOR(l,l,l,1.0f);

			*elem++;
		}
		sceGuDrawArray(GU_TRIANGLES, GU_TEXTURE_32BITF | GU_VERTEX_32BITF | GU_COLOR_8888, numelems, 0, out);

		mesh = (md3mesh_t*)((char*)mesh + mesh->meshsize);
	}
	sceGumPopMatrix();
    sceGumUpdateMatrix();
}

extern char loadname[32];
//extern int particle_tex;
//extern int detailtexture;

byte* COM_LoadFile (char *path, int usehunk) ;

void Mod_LoadQ3Model(model_t *mod, void *buffer)
{
	md3header_t *header;
	md3mesh_t *mesh;
	int i, j, size;
	char *skin;
	int base;
	int *skins;
	int defaultskin;

	char skinname[CLASSIC_MAX_QPATH];
	char  * texnopath;


	int numskins, numskinsfound;
	char *skinsfound[16];


	header = (md3header_t *)buffer;
	mesh = (md3mesh_t*)((byte *)buffer + header->mesh_offs);
	for (size = 0, i = 0; i < header->nummeshes; i++)
	{
		size += mesh->numverts * mesh->numframes * sizeof(vec3_t);
		mesh = (md3mesh_t*)((char*)mesh + mesh->meshsize);
	}

	i = Q_strrchr(mod->name, '/') - mod->name;
	Q_strncpy(skinname, mod->name, i+1);
	Q_strcat(skinname, loadname);
	numskinsfound = COM_MultipleSearch(va("%s*.skin", skinname), skinsfound, 16, qfalse);
	
	Con_DPrintf("Found %d skins files...\n", numskinsfound);

	numskins = numskinsfound;
	if(!numskins)
		numskins = 1;

	size += sizeof(int) * numskins * header->nummeshes;

	if ((int)(mod) && mod->cache.data)
		return;

	header = static_cast<md3header_t*>(Cache_Alloc (&mod->cache, com_filesize + size, loadname));
	if (!mod->cache.data)
		return;

	Q_memcpy(header, buffer, com_filesize);

	base = com_filesize;

	mod->type = mod_md3;//mod_hud...
	//mod->version = header->version;
	mod->numframes = header->numboneframes;
	VectorCopy(((md3boneframe_t *)((byte *)header + header->bone_offs))->mins, mod->mins);
	VectorCopy(((md3boneframe_t *)((byte *)header + header->bone_offs))->maxs, mod->maxs);
	
	mesh = (md3mesh_t *)((byte *)header + header->mesh_offs);

	Con_DPrintf("Procesing  %d meshes...\n", header->nummeshes);
				//Cvar_Set("debug_textures_open","3");

	for (i = 0; i < header->nummeshes; i++)
	{
		md3vert_t *vert;
		if (mesh->numskins != 1)
			Sys_Error("Unexpected number of skins (%i) in %s", mesh->numskins, mesh->name);

		mesh->numskins = numskins;

		skins = (int *)((byte *)header + base);
		if(mesh->skin_offs)
		{
			skin = (char *)mesh + mesh->skin_offs;
			j = Q_strlen(skin);
			if(j)
			{
				int texnum;
				Con_DPrintf("Loading %s skin...\n", skin);
	

				texnum = loadtextureimage(COM_SkipPath(skin),0 ,0 ,qtrue,GU_LINEAR);//TX_LoadTexture(-1, skin, true, 0, 0);
				if (!texnum)
					texnum = loadtextureimage(skin,0 ,0 ,qtrue,GU_LINEAR);;//TX_LoadTexture(-1, skin, true, 0, 0);

				if (!texnum)
					texnum = loadtextureimage(va("progs/%s",COM_SkipPath(skin)),0 ,0 ,qtrue,GU_LINEAR);

				texnopath = COM_SkipPath(skin);

				if (texnum)
					Con_DPrintf("Load texture OK!\n");
				else
					Con_DPrintf("Load texture failed!\n");

				for(j = 0; j < numskins; j++)
					skins[j] = texnum;

                mapTextureNameList.push_back(texnum);
			}
		}
		else
		{
			//skins[0] = particle_tex;//lost texture;
			//for(j = 0; j < numskins; j++)
				//if (particle_tex)
				//	skins[j] = particle_tex;//missign texture
				///else
				//	skins[j] = detailtexture;//missign texture
		}

		mesh->skin_offs = base;
		base += sizeof(int)*numskins;

		vert = (md3vert_t*)((char*)mesh + mesh->vert_offs);
		mesh->vert_offs = base;
		size = mesh->numverts * mesh->numframes;
		base += size * sizeof(vec3_t);
		for (j=0; j < size; j++)
		{
			vec3_t *vertexes = (vec3_t *)((byte *)header + mesh->vert_offs);
			vertexes[j][0] = (float)vert->vec[0] / 64.0f;
			vertexes[j][1] = (float)vert->vec[1] / 64.0f;
			vertexes[j][2] = (float)vert->vec[2] / 64.0f;
			vert++;
		}

		mesh = (md3mesh_t*)((char*)mesh + mesh->meshsize);
	}

	for(i = 0; i < numskinsfound; i++)
		if(strcmp(skinsfound[i], va("%s_default.skin", skinname))==0)
		{
			defaultskin = i;
			break;
		}

	for(i = 0; i < numskinsfound; i++)
	{
		char *token, *oldtoken;
		int pos;
		char *skindata;

		Con_DPrintf("Loading new skin...\n");

		if(!i)
			skindata = (char*)((byte *)(COM_LoadFile(skinsfound[defaultskin], 0)));
		else if(i < defaultskin)
			skindata = (char*)((byte *)(COM_LoadFile(skinsfound[i+1], 0)));
		else
			skindata = (char*)((byte *)(COM_LoadFile(skinsfound[i], 0)));
	
		if(skindata)
		{
			pos = 0;
			while(skindata[pos])
			{
				oldtoken = &skindata[pos];
				token = &skindata[pos];
				while(skindata[pos] != ',' && skindata[pos])
					pos++;
				skindata[pos++] = '\0';

				j = 0;
				mesh = (md3mesh_t *)((byte *)header + header->mesh_offs);
				for(j = 0; j < header->nummeshes && strcmp(mesh->name ,token)!=0; j++)
					mesh = (md3mesh_t*)((char*)mesh + mesh->meshsize);

				token = &skindata[pos];
				while(skindata[pos] != '\n' && skindata[pos])
					pos++;
				skindata[pos++-1] = '\0';

				if(token[0] && j < header->nummeshes)
				{
					skins = (int *)((byte *)header + mesh->skin_offs);
					skins[i] = loadtextureimage(token,0 ,0 ,qfalse,GU_LINEAR);

					if(skins[i] == 0)//missing texture
					{
						int len = Q_strlen(token);

						Con_DPrintf("Loading %s token-skin...\n", token);
						skins[i] = loadtextureimage(token,0 ,0 ,qtrue,GU_LINEAR);

						if (!skins[i])
							skins[i] = loadtextureimage(COM_SkipPath(token),0 ,0 ,qtrue,GU_LINEAR);
						
						if (skins[i])
							Con_DPrintf("Load texture OK skindata!\n");
						else
							Con_DPrintf("Load texture failed skindata!\n");
					}
                    mapTextureNameList.push_back(skins[i]);
				}
			}
			Z_Free(skindata);
		}
		free(skinsfound[i]);
	}
}

