extern "C"
{
#include "../quakedef.h"
}

#include <pspgu.h>
#include <pspgum.h>
#include "video_hardware_md3.h"

#define	INTERP_MAXDIST		300

#define SHADEDOT_QUANT		64 /* like in gl_rmain.c */
#define NUMVERTEXNORMALS	162 /* like in gl_rmain.c */

typedef float m3by3_t[3][3];


extern byte	*shadedots;
extern byte	r_avertexnormal_dots[SHADEDOT_QUANT][NUMVERTEXNORMALS];
extern float	shadelight, ambientlight;
extern vec3_t lightcolor; // LordHavoc: .lit support

int loadtextureimage (char* filename, int matchwidth, int matchheight, qboolean complain, int filter);

extern vec3_t	shadevector;

#include <list>

using namespace std;

extern list<int> mapTextureNameList;

/*
To draw, for each surface, run through the triangles, getting tex coords from s+t, 
*/
void R_DrawQ3Model (entity_t *ent)
{
	float		l, lerpfrac;
	double scale;
	int distance = INTERP_MAXDIST / MD3_XYZ_SCALE;
	vec3_t		interpolated_verts;

	md3model_t *mhead;
	md3Header_t *pheader;
	model_t *mod;
	int surfnum, numtris, i;
	md3Surface_t *surf;
	//Light setups from video_hardware_maim.cpp
	float ang;
	int			lnum;
    vec3_t		dist;
	float		add;

	int frame1 = ent->oldframe, frame2 = ent->frame;
	md3XyzNormal_t *verts, *v1, *v2;

	surfinf_t *sinf;

	unsigned int	*tris;
	md3St_t *tc;

	mod = ent->model;

   if(ISADDITIVE(ent))
   {
	 float deg = ent->renderamt / 255.0f;
	 float alpha_val  = deg;
     float alpha_val2 = 1 - deg;
	 if(deg <= 0.7)
	 sceGuDepthMask(GU_TRUE);

	 sceGuEnable (GU_BLEND);
	 sceGuBlendFunc(GU_ADD, GU_FIX, GU_FIX,
	 GU_COLOR(alpha_val,alpha_val,alpha_val,alpha_val),
	 GU_COLOR(alpha_val2,alpha_val2,alpha_val2,alpha_val2));
   }
   else if(ISGLOW(ent))
   {
	 sceGuDepthMask(GU_TRUE);
	 sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_FIX, 0, 0xFFFFFFFF);
	 sceGuTexFunc(GU_TFX_MODULATE , GU_TCC_RGB);
   }
   else if (ISSOLID(ent))
   {
	 sceGuEnable(GU_ALPHA_TEST);
	 float c = (ent->renderamt) * 1.0f;
	 sceGuAlphaFunc(GU_GREATER, 0x88, c);
   }

	R_LightPoint(ent->origin); // LordHavoc: lightcolor is all that matters from this

	for (lnum=0 ; lnum<MAX_DLIGHTS ; lnum++)
	{
		if (cl_dlights[lnum].die >= cl.time)
		{
			VectorSubtract (currententity->origin,
							cl_dlights[lnum].origin,
							dist);
			add = cl_dlights[lnum].radius - Length(dist);

			// LordHavoc: .lit support begin
			/* LordHavoc: original code
			if (add > 0) {
				ambientlight += add;
				//ZOID models should be affected by dlights as well
				shadelight += add;
			}
			*/
			if (add > 0)
			{
				lightcolor[0] += add * cl_dlights[lnum].color[0];
				lightcolor[1] += add * cl_dlights[lnum].color[1];
				lightcolor[2] += add * cl_dlights[lnum].color[2];
			}
			// LordHavoc: .lit support end
		}
	}

	if (lightcolor[0] < 8)
		lightcolor[0] = 8;

	if (lightcolor[1] < 8)
		lightcolor[1] = 8;

	if (lightcolor[2] < 8)
		lightcolor[2] = 8;

	if (lightcolor[0] > 125)
	    lightcolor[0] = 125;

	if (lightcolor[1] > 125)
        lightcolor[1] = 125;

    if (lightcolor[2] > 125)
        lightcolor[2] = 125;


	shadedots = r_avertexnormal_dots[((int) (ent->angles[1] * (SHADEDOT_QUANT / 360.0))) & (SHADEDOT_QUANT - 1)];

	VectorScale(lightcolor, 1.0f / 200.0f, lightcolor);


	ang = ent->angles[1]/180*M_PI;
	shadevector[0] = cos(-ang);
	shadevector[1] = sin(-ang);
	shadevector[2] = 1;
	VectorNormalize (shadevector);


	sceGumPushMatrix ();
	R_RotateForEntity (ent);
	scale = 1;
	// perform two scalling at once, one scalling for MD3_XYZ_SCALE, other for r_viewmodelsize
    // Scale.
    const ScePspFVector3 GUscale =
	{
	scale * MD3_XYZ_SCALE, MD3_XYZ_SCALE, MD3_XYZ_SCALE
	};
	sceGumScale(&GUscale);

	sceGuShadeModel(GU_SMOOTH);

	sceGumUpdateMatrix();

   	mhead = (md3model_t *)Mod_Extradata (mod);
	sinf = (surfinf_t *)((char *)mhead + mhead->surfinf);
	pheader = (md3Header_t *)((char *)mhead + mhead->md3model);

	if (frame1 >= pheader->numFrames)
		frame1 = pheader->numFrames-1;
	if (frame2 >= pheader->numFrames)
		frame2 = pheader->numFrames-1;

	//if (!r_lerpframes.value || ent->framelerp < 0 || ent->oldframe == ent->frame)
	//	lerpfrac = 1.0;
	//else
		lerpfrac = fmin (ent->framelerp, 1);

	surf = (md3Surface_t *)((char *)pheader + pheader->ofsSurfaces);

	for (surfnum = 0; surfnum < pheader->numSurfaces; surfnum++) //loop through the surfaces.
	{
		int pose1, pose2;

		pose1 = frame1*surf->numVerts;
		pose2 = frame2*surf->numVerts;

		tc = (md3St_t *)((char *)surf + surf->ofsSt);	//skin texture coords.
		verts = (md3XyzNormal_t *)((char *)surf + surf->ofsXyzNormals);

		tris = (unsigned int *)((char *)surf + surf->ofsTriangles);
		numtris = surf->numTriangles * 3;		

		GL_Bind((sinf+pheader->numSurfaces*pheader->numSkins + surfnum)->texnum);

		// Allocate the vertices.
		struct vertex
		{
			float u, v;
			unsigned int color;
			float x, y, z;
		};

		vertex* const out = static_cast<vertex*>(sceGuGetMemory(sizeof(vertex) * numtris));

		for (i = 0 ; i < numtris ; i++)
		{
			float	s, t;

			v1 = verts + *tris + pose1;
			v2 = verts + *tris + pose2;

			s = tc[*tris].s, t = tc[*tris].t;

			out[i].u = s;
			out[i].v = t;
    
			lerpfrac = VectorL2Compare(v1->xyz, v2->xyz, distance) ? lerpfrac : 1;

			//Light setup
            //l = FloatInterpolate (shadedots[v1->normal>>8], lerpfrac, shadedots[v2->normal>>8]);
			//l = (l * shadelight + ambientlight) / 256;
			//l = fmin(l, 1);

			l = shadedots[v1->normal>>8];

			float       r,g,b;

			r = l * lightcolor[0];
            g = l * lightcolor[1];
            b = l * lightcolor[2];

			if (r > 1.0f)
                r = 1.0f;
            if (g > 1.0f)
                g = 1.0f;
            if (b > 1.0f)
                b = 1.0f;

			VectorInterpolate (v1->xyz, lerpfrac, v2->xyz, interpolated_verts);

			out[i].x = interpolated_verts[0];
			out[i].y = interpolated_verts[1];
            out[i].z = interpolated_verts[2];

			out[i].color = GU_COLOR(r, g, b, 1.0f);

			tris++;
		}
        sceGuDrawArray(GU_TRIANGLES, GU_TEXTURE_32BITF | GU_VERTEX_32BITF | GU_COLOR_8888, numtris, 0, out);

		surf = (md3Surface_t *)((char *)surf + surf->ofsEnd);	//NEXT!   Getting cocky!
	}

	if (ISADDITIVE(ent))
    {
        float deg = ent->renderamt / 255.0f;

		if(deg <= 0.7)
		   sceGuDepthMask(GU_FALSE);

		sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
        sceGuDisable (GU_BLEND);
    }
    else if(ISSOLID(ent))
    {
	    sceGuAlphaFunc(GU_GREATER, 0, 0xff);
	    sceGuDisable(GU_ALPHA_TEST);
    }
    else if(ISGLOW(ent))
    {
        sceGuDepthMask(GU_FALSE);
		sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
        sceGuDisable (GU_BLEND);
    }

	sceGuShadeModel(GU_FLAT);
	sceGumPopMatrix();
	sceGumUpdateMatrix();
	sceGuEnable(GU_TEXTURE_2D);
}

int Mod_ReadFlagsFromMD1(char *name, int md3version)
{
	mdl_t				*pinmodel;
	char fname[MAX_QPATH];
	COM_StripExtension(name, fname);
	COM_DefaultExtension(fname, ".md3");
	
	if (!strcmp(name, fname))	//md3 renamed as mdl
	{
		COM_StripExtension(name, fname);	//seeing as the md3 is named over the mdl,
		COM_DefaultExtension(fname, ".md1");//read from a file with md1 (one, not an ell)
	}
	else
	{
		COM_StripExtension(name, fname);
		COM_DefaultExtension(fname, ".mdl");
	}

	pinmodel = (mdl_t *)COM_LoadTempFile(fname);


	if (!pinmodel)	//not found
		return 0;

	if (pinmodel->ident != IDPOLYHEADER)
		return 0;
	if (pinmodel->version != ALIAS_VERSION)
		return 0;

	return pinmodel->flags;
}

byte *FS_LoadFile (char *path, int usehunk); //for mem allocation

void Mod_LoadQ3Model (model_t *mod, void *buffer)
{
#define ll(x) x=LittleLong(x)	//easier to type byte swap
#define lf(x) x=LittleFloat(x)
	char specifiedskinname[128];
	char skinfileskinname[128];
	char tenebraeskinname[128];
	extern char loadname[];
	int					start, end, total;
	int numsurfs;
	int numskins;
	int surfn;

	md3model_t *pheader;
	md3Header_t *mem;
	surfinf_t *sinf;
	md3Surface_t *surf;
	md3Shader_t *sshad;
	md3Frame_t *mFrame;
	int fr, i, j;
	md3St_t *st;
	int		size;

	start = Hunk_LowMark ();

	numsurfs = LittleLong(((md3Header_t *)buffer)->numSurfaces);
	numskins = 1;

	//Crow_bar. memory cached fixed
	size = LittleLong(((md3Header_t *)buffer)->ofsEnd);

	pheader = (md3model_t *) Hunk_Alloc(sizeof(md3model_t) +(numsurfs*numskins)*sizeof(surfinf_t));
	pheader->surfinf = sizeof(md3model_t);

	mem = (md3Header_t *) Hunk_Alloc(size);
    //===

	//pointer to header

	pheader->md3model = (char *)mem - (char *)pheader;

	memcpy(mem, buffer, size);	//casually load the entire thing. As you do.

	ll(mem->ident);
	ll(mem->version);

	ll(mem->flags);	//Does anyone know what these are?

	ll(mem->numFrames);
	ll(mem->numTags);
	ll(mem->numSurfaces);

	ll(mem->numSkins);

	ll(mem->ofsFrames);
	ll(mem->ofsTags);
	ll(mem->ofsSurfaces);
	ll(mem->ofsEnd);


	pheader->numframes = mem->numFrames;
	pheader->numtags = mem->numTags;
	pheader->ofstags = pheader->md3model + mem->ofsTags;	
	for (fr = 0; fr < mem->numFrames; fr++)
	{
		mFrame = ((md3Frame_t *)((char *)mem + mem->ofsFrames)) + fr;
		for (j = 0; j < 3; j++)
		{
			lf(mFrame->bounds[0][j]);
			lf(mFrame->bounds[1][j]);
			lf(mFrame->localOrigin[j]);
		}
		ll(mFrame->radius);
	}

	sinf = (surfinf_t*)((char *)pheader + pheader->surfinf);
//	for (skinn = 0; skinn < numskins; skinn++)
	{
		surf = (md3Surface_t *)((char *)mem + mem->ofsSurfaces);
		for (surfn = 0; surfn < numsurfs; surfn++)
		{
			md3Triangle_t	*tris;
			md3XyzNormal_t	*vert;

			ll(surf->ident);

			ll(surf->flags);
			ll(surf->numFrames);

			ll(surf->numShaders);
			ll(surf->numVerts);

			ll(surf->numTriangles);
			ll(surf->ofsTriangles);

			ll(surf->ofsShaders);
			ll(surf->ofsSt);
			ll(surf->ofsXyzNormals);

			ll(surf->ofsEnd);

			st = (md3St_t *)((char *)surf + surf->ofsSt);	//skin texture coords.
			for (i = 0; i < surf->numVerts; i++)
			{
				lf(st[i].s);
				lf(st[i].t);
			}

			// swap all the triangles
			tris = (md3Triangle_t *)((char *)surf + surf->ofsTriangles);
			for (j=0 ; j<surf->numTriangles ; j++)
			{
				ll(tris[j].indexes[0]);
				ll(tris[j].indexes[1]);
				ll(tris[j].indexes[2]);
			}

			// swap all the vertices
			vert = (md3XyzNormal_t *)((char *)surf + surf->ofsXyzNormals);
			for (j=0 ; j < surf->numVerts * surf->numFrames ; j++)
			{
				vert[j].xyz[0] = LittleShort (vert[j].xyz[0]);
				vert[j].xyz[1] = LittleShort (vert[j].xyz[1]);
				vert[j].xyz[2] = LittleShort (vert[j].xyz[2]);
				vert[j].normal = LittleShort (vert[j].normal);
			}

			sshad = (md3Shader_t *)((char *)surf + surf->ofsShaders);

			ll(sshad->shaderIndex);

			*specifiedskinname = *skinfileskinname = *tenebraeskinname = '\0';
			strlcpy (specifiedskinname, sshad->name, sizeof (specifiedskinname));

			if (*sshad->name)
			{		
				strlcpy (tenebraeskinname, mod->name, sizeof (tenebraeskinname)); //backup
				strcpy (COM_SkipPathWritable(tenebraeskinname), sshad->name);
			}
			else
			{
				char *sfile, *sfilestart, *nl;
				int len;

				//hmm. Look in skin file.
				strlcpy (sinf->name, mod->name, sizeof (sinf->name));
				COM_StripExtension(sinf->name, sinf->name);
				strlcat (sinf->name, "_default.skin", sizeof (sinf->name));

				sfile = sfilestart = (char *) COM_LoadHunkFile(sinf->name);

				strlcpy (sinf->name, mod->name, sizeof (sinf->name)); //backup
				COM_StripExtension(sinf->name, sinf->name);
				strlcat (sinf->name, "_skin.tga", sizeof (sinf->name));

				len = strlen(surf->name);

				if (sfilestart)
				{
					while(*sfile)
					{
						nl = strchr(sfile, '\n');
						if (!nl)
							nl = sfile + strlen(sfile);
						if (sfile[len] == ',' && !strncasecmp(surf->name, sfile, len))
						{
							strlcpy(skinfileskinname, sfile+len+1, nl - (sfile + len) -2);
							break;
						}
						sfile = nl+1;
					}
					//Z_Free(sfilestart);
				}
				
			}
			//now work out which alternative is best, and load it.
			if (*skinfileskinname && (sinf->texnum=loadtextureimage(skinfileskinname, 0, 0, qtrue, GU_LINEAR)))
				strlcpy (sinf->name, skinfileskinname, sizeof (sinf->name));
			else if (*specifiedskinname && (sinf->texnum=loadtextureimage(specifiedskinname, 0, 0, qtrue, GU_TRUE)))
				strlcpy (sinf->name, specifiedskinname, sizeof (sinf->name));
			else if (*tenebraeskinname)
			{
				sinf->texnum=loadtextureimage(tenebraeskinname, 0, 0, qtrue, GU_LINEAR);
				strlcpy (sinf->name, tenebraeskinname, sizeof (sinf->name));
			}
			else if (*skinfileskinname)
			{
				sinf->texnum=loadtextureimage(tenebraeskinname, 0, 0, qtrue, GU_LINEAR);
				strlcpy (sinf->name, skinfileskinname, sizeof (sinf->name));
			}
			else
			{
				sinf->texnum=loadtextureimage("dummy", 0, 0, qfalse, GU_LINEAR);
				strlcpy (sinf->name, "dummy", sizeof (sinf->name));
			}

			mapTextureNameList.push_back(sinf->texnum);

			surf = (md3Surface_t *)((char *)surf + surf->ofsEnd);

			sinf++;
		}
	}

	mod->type = mod_md3;

	end = Hunk_LowMark ();
	total = end - start;


	Cache_Alloc (&mod->cache, total, loadname);

	if (!mod->cache.data)
		return;

	memcpy (mod->cache.data, pheader, total);

	Hunk_FreeToLowMark (start);

	mod->flags = Mod_ReadFlagsFromMD1(mod->name, mem->flags);
}


int GetTag(model_t *mod, char *tagname, int frame, float **org, m3by3_t **ang)	//if anyone is interested that is...
{
	md3model_t *mhead;
	md3tag_t *tag;
	int tnum;
	float bad[3]={0};
	m3by3_t badm= {{0}};

	*org = bad;
	*ang = &badm;

	Con_Printf("TAG S - OK\n");

	if (mod->type != mod_md3)
		return false;

	mhead = (md3model_t *)Mod_Extradata (mod);
	if (frame > mhead->numframes)
		frame = 0;	

	tag = (md3tag_t *)((char *)mhead + mhead->ofstags);

	tag += frame*mhead->numtags;
	
	for(tnum=0;tnum<mhead->numtags; tnum++)
	{
		if (!strcasecmp(tag->name, tagname))
		{
			*org = tag->org;
			*ang = &tag->ang;
			return true;
		}
		tag++;
	}

    Con_Printf("TAG E - OK\n");

	return false;
}
