#include <pspkernel.h>
#include <pspdebug.h>
#include <pspdisplay.h>
#include <pspctrl.h>
#include <pspaudiolib.h>
#include <pspaudio.h>
#include <psputility.h>
#include <pspmp3.h>

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include "oggplayer.h"
#include <pspgu.h>
#include <pspgum.h>
#include <mikmod.h>
#include "font.c"

/*
PSP_MODULE_INFO("mediamodule", 0, 0, 0);
PSP_MAIN_THREAD_ATTR(0);
*/
PSP_MODULE_INFO("mediamodule", PSP_MODULE_USER, 1, 0);
PSP_MAIN_THREAD_ATTR(PSP_THREAD_ATTR_USER | PSP_THREAD_ATTR_VFPU);
PSP_HEAP_SIZE_KB(-1024);

void play_ogg(char *file);
int exists(char *file);
void get_fileext(char *file, char *ext);
int is_mikmodfile(char *ext);

#define printf	pspDebugScreenPrintf

static unsigned int __attribute__((aligned(16))) list[262144];

static int fontwidthtab[128] = {
	10, 10, 10, 10,
	10, 10, 10, 10,
	10, 10, 10, 10,
	10, 10, 10, 10,

	10, 10, 10, 10,
	10, 10, 10, 10,
	10, 10, 10, 10,
	10, 10, 10, 10,

	10,  6,  8, 10, //   ! " #
	10, 10, 10,  6, // $ % & '
	10, 10, 10, 10, // ( ) * +
	 6, 10,  6, 10, // , - . /

	10, 10, 10, 10, // 0 1 2 3
	10, 10, 10, 10, // 6 5 8 7
	10, 10,  6,  6, // 10 9 : ;
	10, 10, 10, 10, // < = > ?

	16, 10, 10, 10, // @ A B C
	10, 10, 10, 10, // D E F G
	10,  6,  8, 10, // H I J K
	 8, 10, 10, 10, // L M N O

	10, 10, 10, 10, // P Q R S
	10, 10, 10, 12, // T U V W
	10, 10, 10, 10, // X Y Z [
	10, 10,  8, 10, // \ ] ^ _

	 6,  8,  8,  8, // ` a b c
	 8,  8,  6,  8, // d e f g
	 8,  6,  6,  8, // h i j k
	 6, 10,  8,  8, // l m n o

	 8,  8,  8,  8, // p q r s
	 8,  8,  8, 12, // t u v w
	 8,  8,  8, 10, // x y z {
	 8, 10,  8, 12  // | } ~
};

#define BUF_WIDTH (512)
#define SCR_WIDTH (480)
#define SCR_HEIGHT (272)
#define PIXEL_SIZE (4)
#define FRAME_SIZE (BUF_WIDTH * SCR_HEIGHT * PIXEL_SIZE)
#define ZBUF_SIZE (BUF_WIDTH SCR_HEIGHT * 2)

static void SetupGu()
{
    sceGuInit();
	sceGuStart(GU_DIRECT, list);
	sceGuDrawBuffer(GU_PSM_8888,(void*)0,BUF_WIDTH);
	sceGuDispBuffer(SCR_WIDTH,SCR_HEIGHT,(void*)0x88000,BUF_WIDTH);
	sceGuDepthBuffer((void*)0x110000,BUF_WIDTH);
	sceGuOffset(2048 - (SCR_WIDTH/2),2048 - (SCR_HEIGHT/2));
	sceGuViewport(2048,2048,SCR_WIDTH,SCR_HEIGHT);
	sceGuDepthRange(0xc350,0x2710);
	sceGuScissor(0,0,SCR_WIDTH,SCR_HEIGHT);
	sceGuEnable(GU_SCISSOR_TEST);
	sceGuDisable(GU_DEPTH_TEST);
	sceGuShadeModel(GU_SMOOTH);
	sceGuEnable(GU_BLEND);
	sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
	sceGuEnable(GU_TEXTURE_2D);
	sceGuTexMode(GU_PSM_8888, 0, 0, 0);
	sceGuTexImage(0, 256, 128, 256, font);
	sceGuTexFunc(GU_TFX_MODULATE, GU_TCC_RGBA);
	sceGuTexEnvColor(0x0);
	sceGuTexOffset(0.0f, 0.0f);
	sceGuTexScale(1.0f / 256.0f, 1.0f / 128.0f);
	sceGuTexWrap(GU_REPEAT, GU_REPEAT);
	sceGuTexFilter(GU_NEAREST, GU_NEAREST);
	sceGuFinish();
	sceGuSync(0,0);
	sceGuDisplay(GU_TRUE);
}

static void ShGu()
{
	sceGuDisplay(GU_FALSE);
	sceGuTerm();
}

void drawString(const char* text, int x, int y, unsigned int color, int fw) {
	int len = (int)strlen(text);
	if(!len) {
		return;
	}

	typedef struct {
		float s, t;
		unsigned int c;
		float x, y, z;
	} VERT;

	VERT* v = sceGuGetMemory(sizeof(VERT) * 2 * len);

	int i;
	for(i = 0; i < len; i++) {
		unsigned char c = (unsigned char)text[i];
		if(c < 32) {
			c = 0;
		} else if(c >= 128) {
			c = 0;
		}

		int tx = (c & 0x0F) << 4;
		int ty = (c & 0xF0);

		VERT* v0 = &v[i*2+0];
		VERT* v1 = &v[i*2+1];

		v0->s = (float)(tx + (fw ? ((16 - fw) >> 1) : ((16 - fontwidthtab[c]) >> 1)));
		v0->t = (float)(ty);
		v0->c = color;
		v0->x = (float)(x);
		v0->y = (float)(y);
		v0->z = 0.0f;

		v1->s = (float)(tx + 16 - (fw ? ((16 - fw) >> 1) : ((16 - fontwidthtab[c]) >> 1)));
		v1->t = (float)(ty + 16);
		v1->c = color;
		v1->x = (float)(x + (fw ? fw : fontwidthtab[c]));
		v1->y = (float)(y + 16);
		v1->z = 0.0f;

		x += (fw ? fw : fontwidthtab[c]);
	}

	sceGumDrawArray(GU_SPRITES,
		GU_TEXTURE_32BITF | GU_COLOR_8888 | GU_VERTEX_32BITF | GU_TRANSFORM_2D,
		len * 2, 0, v
	);
}

// Input and Output buffers
char	mp3Buf[16*1024]  __attribute__((aligned(64)));
short	pcmBuf[16*(1152/2)]  __attribute__((aligned(64)));

int fillStreamBuffer( int fd, int handle )
{
	char* dst;
	int write;
	int pos;
	// Get Info on the stream (where to fill to, how much to fill, where to fill from)
	int status = sceMp3GetInfoToAddStreamData( handle, &dst, &write, &pos);
	if (status<0)
	{
		printf("ERROR: sceMp3GetInfoToAddStreamData returned 0x%08X\n", status);
	}

	// Seek file to position requested
	status = sceIoLseek32( fd, pos, SEEK_SET );
	if (status<0)
	{
		printf("ERROR: sceIoLseek32 returned 0x%08X\n", status);
	}

	// Read the amount of data
	int read = sceIoRead( fd, dst, write );
	if (read < 0)
	{
		printf("ERROR: Could not read from file - 0x%08X\n", read);
	}

	if (read==0)
	{
		// End of file?
		return 0;
	}

	// Notify mp3 library about how much we really wrote to the stream buffer
	status = sceMp3NotifyAddStreamData( handle, read );
	if (status<0)
	{
		printf("ERROR: sceMp3NotifyAddStreamData returned 0x%08X\n", status);
	}

	return (pos>0);
}

/* Exit callback */
int exitCallback(int arg1, int arg2, void *common)
{
	sceKernelExitGame();
	return 0;
}

/* Callback thread */
int callbackThread(SceSize args, void *argp)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", (void*) exitCallback, NULL);
	sceKernelRegisterExitCallback(cbid);
	sceKernelSleepThreadCB();

	return 0;
}

/* Sets up the callback thread and returns its thread id */
int setupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", callbackThread, 0x11, 0xFA0, 0, 0);
	if (thid >= 0) {
		sceKernelStartThread(thid, 0, 0);
	}
	return thid;
}

int exists(char *file)
{
	int x;
	if((x=sceIoOpen(file, PSP_O_RDONLY, 0777))>=0)
	   return (sceIoClose(x) ? 1 : 1);
	else return 0;
}

void get_fileext(char *file, char *ext)
{
	char *f=file;
	
	while(strstr(f, ".")) f=strstr(f, ".")+1;
	strcpy(ext,f);
}

#define NUM_MIKMODFORMATS 20
char *mikmodformats[NUM_MIKMODFORMATS] =
{
	"mod",
	"669",
	"amf",
	"apun",
	"dsm",
	"far",
	"gdm",
	"it",
	"imf",
	"mod",
	"med",
	"mtm",
	"okt",
	"s3m",
	"stm",
	"stx",
	"ult",
	"uni",
	"xm",
    "umx" //test
};

int is_mikmodfile(char *ext)
{
	int x, y;
	for(x=0;x<NUM_MIKMODFORMATS;x++)
		for(y=0;y<strlen(mikmodformats[x]);y++)
			if(ext[y]==mikmodformats[x][y] || ext[y]==mikmodformats[x][y]-32) return 1;

	return 0;
}

static int oldButtons = 0;
int changedButtons=0;
SceCtrlData pad;

void play_mikmod(char *file)
{
	if(!exists(file))
	{
		while(1)
		{
			sceGuStart(GU_DIRECT, list);
		    sceGuClear(GU_COLOR_BUFFER_BIT);

			char a[80];
			sprintf(a,"\n%s not found!", file);
            drawString(a,  0, 0, 0xFF0000FF, 0);

			sceGuFinish();
		    sceGuSync(0, 0);

		    sceDisplayWaitVblankStart();
		    sceGuSwapBuffers();
		}
		return;
	}
	
	FILE *modfile = fopen(file,"rb");
	MODULE *mod = Player_LoadFP(modfile, 80, 0);
	if(mod)
	{
     	Player_Start(mod);
		while(Player_Active() && !(changedButtons & PSP_CTRL_CROSS))
		{
			char a[80],c[80],d[80],e[80],f[80];

			sceGuStart(GU_DIRECT, list);
		    sceGuClear(GU_COLOR_BUFFER_BIT);

			static float t = 0.0f;
		    t += 0.1f;

		    unsigned int h = 0xFF000000 |
			 (unsigned int)((sinf(t * 0.393f + 0.086f) / 2.0f + 0.5f) * 255.0f) << 16 |
			 (unsigned int)((sinf(t * 0.444f + 0.854f) / 2.0f + 0.5f) * 255.0f) <<  8 |
			 (unsigned int)((sinf(t * 0.117f + 1.337f) / 2.0f + 0.5f) * 255.0f) <<  0;
			sprintf(f,"Mod Player by Crow_bar for IRSHELL 5.xx\n");
            drawString(f, 0, 16 * 0, h, 0);

			sprintf(a,"Playing '%s'\n", file);
            drawString(a, 0, 16 * 1, 0xFF00FF00, 0);

			drawString("play/pause: start\n", 0, 16 * 2, 0xFF0000FF, 0);
            drawString("stop: cross\n", 0, 16 * 3, 0xFF0000FF, 0);
            drawString("previous/next position: d-pad left/right\n", 0, 16 * 4, 0xFF0000FF, 0);
            drawString("start over: square\n", 0, 16 * 5, 0xFF0000FF, 0);

			if(mod->songname)
			{
			  sprintf(c,"Song Title: %s\n", mod->songname);
              drawString(c, 0, 16 * 6, 0xFFFF0000, 0);
			}
			if(mod->comment)
		    {
              unsigned int j = 0xFFFF0000 |
			  (unsigned int)((sinf(t * 0.393f + 0.086f) / 2.0f + 0.5f) * 255.0f) << 16 |
			  (unsigned int)((sinf(t * 0.444f + 0.854f) / 2.0f + 0.5f) * 255.0f) <<  8 |
			  (unsigned int)((sinf(t * 0.117f + 1.337f) / 2.0f + 0.5f) * 255.0f) <<  0;
			  sprintf(d,"Song Comment: %s\n", mod->comment);
              drawString(d, 0, 16 * 7, j, 0);
			}

			sceCtrlReadBufferPositive(&pad, 1);
			
			sprintf(e,"Current position: %i of %i              ", mod->sngpos, mod->numpos);
            drawString(e, 0, 16 * 8, 0xFFFFFF00, 0);

			changedButtons = pad.Buttons & (~oldButtons);
			MikMod_Update();
			if (changedButtons & PSP_CTRL_START)
				Player_TogglePause();
			else if(changedButtons & PSP_CTRL_SQUARE)
				Player_SetPosition(mod->reppos);
			else if(changedButtons & PSP_CTRL_LEFT)
				Player_PrevPosition();
			else if(changedButtons & PSP_CTRL_RIGHT)
				Player_NextPosition();
			oldButtons = pad.Buttons;

			sceGuFinish();
		    sceGuSync(0, 0);

		    sceDisplayWaitVblankStart();
		    sceGuSwapBuffers();
		}
		Player_Stop();
		Player_Free(mod);
	}
	else
	{
        int t = 0;
		while(t < 100) //~3 sec
		{
			sceGuStart(GU_DIRECT, list);
		    sceGuClear(GU_COLOR_BUFFER_BIT);

            drawString("Error loading module!",  0, 0, 0xFF0000FF, 0);

			sceGuFinish();
		    sceGuSync(0, 0);

		    sceDisplayWaitVblankStart();
		    sceGuSwapBuffers();
			t++;
		}
	}
}

void play_mp3(char *file)
{
	if(!exists(file))
	{
		int t = 0;
		while(t < 100) //~3 sec
		{
			sceGuStart(GU_DIRECT, list);
		    sceGuClear(GU_COLOR_BUFFER_BIT);

			char a[80];
			sprintf(a,"\n%s not found!", file);
            drawString(a,  0, 0, 0xFF0000FF, 0);

			sceGuFinish();
		    sceGuSync(0, 0);

		    sceDisplayWaitVblankStart();
		    sceGuSwapBuffers();
            t++;
		}
		return;
	}

	// Load modules
	int status = sceUtilityLoadModule(PSP_MODULE_AV_AVCODEC);
	if (status<0)
	{
		printf("ERROR: sceUtilityLoadModule(PSP_MODULE_AV_AVCODEC) returned 0x%08X\n", status);
	}

	status = sceUtilityLoadModule(PSP_MODULE_AV_MP3);
	if (status<0)
	{
		printf("ERROR: sceUtilityLoadModule(PSP_MODULE_AV_MP3) returned 0x%08X\n", status);
	}

	// Open the input file
	int fd = sceIoOpen(file, PSP_O_RDONLY, 0777 );
	if (fd<0)
	{
		printf("ERROR: Could not open file '%s' - 0x%08X\n", file, fd);
	}

	// Init mp3 resources
	status = sceMp3InitResource();
	if (status<0)
	{
		printf("ERROR: sceMp3InitResource returned 0x%08X\n", status);
	}

	// Reserve a mp3 handle for our playback
	SceMp3InitArg mp3Init;
	mp3Init.mp3StreamStart = 0;
	mp3Init.mp3StreamEnd = sceIoLseek32( fd, 0, SEEK_END );
	mp3Init.unk1 = 0;
	mp3Init.unk2 = 0;
	mp3Init.mp3Buf = mp3Buf;
	mp3Init.mp3BufSize = sizeof(mp3Buf);
	mp3Init.pcmBuf = pcmBuf;
	mp3Init.pcmBufSize = sizeof(pcmBuf);

	int handle = sceMp3ReserveMp3Handle( &mp3Init );
	if (handle<0)
	{
		printf("ERROR: sceMp3ReserveMp3Handle returned 0x%08X\n", handle);
	}

	// Fill the stream buffer with some data so that sceMp3Init has something to work with
	fillStreamBuffer( fd, handle );

	status = sceMp3Init( handle );
	if (status<0)
	{
		printf("ERROR: sceMp3Init returned 0x%08X\n", status);
	}

	int channel = -1;
	int samplingRate = sceMp3GetSamplingRate( handle );
	int numChannels = sceMp3GetMp3ChannelNum( handle );
	int lastDecoded = 0;
	int volume = PSP_AUDIO_VOLUME_MAX;
	int numPlayed = 0;
	int paused = 0;
	int lastButtons = 0;
	int loop = 0;
	while (1)
	{
		char a[80],b[80],c[80],d[80],e[80],f[80],g[80];

		sceGuStart(GU_DIRECT, list);
		sceGuClear(GU_COLOR_BUFFER_BIT);

		sprintf(a,"Mod Player by Crow_bar for IRSHELL 5.xx\n");
		sprintf(b,"Playing '%s'\n", file);
		sprintf(c," %i Hz\n", samplingRate);
		sprintf(d," %i kbit/s\n", sceMp3GetBitRate( handle ));
		sprintf(e," %s\n", numChannels==2?"Stereo":"Mono");
		sprintf(f," %s\n", loop==0?"No loop":"Loop");
		int playTime = samplingRate>0?numPlayed / samplingRate:0;
		sprintf(g,"Playtime: %02i:%02i\n", playTime/60, playTime%60 );

        static float t = 0.0f;
		t += 0.1f;

		unsigned int h = 0xFF000000 |
			 (unsigned int)((sinf(t * 0.393f + 0.086f) / 2.0f + 0.5f) * 255.0f) << 16 |
			 (unsigned int)((sinf(t * 0.444f + 0.854f) / 2.0f + 0.5f) * 255.0f) <<  8 |
			 (unsigned int)((sinf(t * 0.117f + 1.337f) / 2.0f + 0.5f) * 255.0f) <<  0;

		drawString(a, 0, 16 * 0, h,          0);
        drawString(b, 0, 16 * 1, 0xFF00FF00, 0);
        drawString(c, 0, 16 * 2, 0xFFFF0000, 0);
        drawString(d, 0, 16 * 3, 0xFFFF0000, 0);
        drawString(e, 0, 16 * 4, 0xFFFF0000, 0);
        drawString(f, 0, 16 * 5, 0xFFFF0000, 0);
        
		drawString("Press CIRCLE to Pause/Resume playback\n", 0, 16 * 6, 0xFF0000FF, 0);
        drawString("Press TRIANGLE to reset playback\n",      0, 16 * 7, 0xFF0000FF, 0);
		drawString("Press SQUARE to switch loop mode\n",      0, 16 * 8, 0xFF0000FF, 0);
		drawString("Press CROSS to stop playback and quit\n", 0, 16 * 9, 0xFF0000FF, 0);
        drawString(g, 0, 16 * 10, 0xFFFFFF00, 0);

		if (!paused)
		{
			// Check if we need to fill our stream buffer
			if (sceMp3CheckStreamDataNeeded( handle )>0)
			{
				fillStreamBuffer( fd, handle );
			}

			// Decode some samples
			short* buf;
			int bytesDecoded;
			int retries = 0;
			// We retry in case it's just that we reached the end of the stream and need to loop
			for (;retries<1;retries++)
			{
				bytesDecoded = sceMp3Decode( handle, &buf );
				if (bytesDecoded>0)
					break;

				if (sceMp3CheckStreamDataNeeded( handle )<=0)
					break;

				if (!fillStreamBuffer( fd, handle ))
				{
					numPlayed = 0;
				}
			}
			if (bytesDecoded<0 && bytesDecoded!=0x80671402)
			{
				printf("ERROR: sceMp3Decode returned 0x%08X\n", bytesDecoded);
			}

			// Nothing more to decode? Must have reached end of input buffer
			if (bytesDecoded==0 || bytesDecoded==0x80671402)
			{
				paused = 1;
				sceMp3ResetPlayPosition( handle );
				numPlayed = 0;
			}
			else
			{
				// Reserve the Audio channel for our output if not yet done
				if (channel<0 || lastDecoded!=bytesDecoded)
				{
					if (channel>=0)
						sceAudioSRCChRelease();

					channel = sceAudioSRCChReserve( bytesDecoded/(2*numChannels), samplingRate, numChannels );
				}
				// Output the decoded samples and accumulate the number of played samples to get the playtime
				numPlayed += sceAudioSRCOutputBlocking( volume, buf );
			}
		}

		sceCtrlPeekBufferPositive(&pad, 1);

		if (pad.Buttons!=lastButtons)
		{
			if (pad.Buttons & PSP_CTRL_CIRCLE)
			{
				paused ^= 1;
			}

			if (pad.Buttons & PSP_CTRL_TRIANGLE)
			{
				// Reset the stream and playback status
				sceMp3ResetPlayPosition( handle );
				numPlayed = 0;
			}

			if (pad.Buttons & PSP_CTRL_SQUARE)
			{
				loop = (loop==0?-1:0);
				status = sceMp3SetLoopNum( handle, loop );
				if (status<0)
				{
					char a[80];
					sprintf(a,"ERROR: sceMp3SetLoopNum returned 0x%08X\n", status);
                    drawString(a, 0, 16 * 11, 0xFF0000FF, 0);
				}
			}

			if (pad.Buttons & PSP_CTRL_CROSS)
			{
				break;
			}

			lastButtons = pad.Buttons;
		}

		sceGuFinish();
		sceGuSync(0, 0);

		sceDisplayWaitVblankStart();
		sceGuSwapBuffers();
	}

    // Cleanup time...
    if (channel>=0)
		sceAudioSRCChRelease();

	status = sceMp3ReleaseMp3Handle( handle );
	status = sceMp3TermResource();
	status = sceIoClose( fd );
}

int was_ogg = 0;

void play_ogg(char *file)
{
	char ogginfo[1024];
	if(!exists(file))
	{
        int t = 0;
		while(t < 100) //~3 sec
		{
			sceGuStart(GU_DIRECT, list);
		    sceGuClear(GU_COLOR_BUFFER_BIT);

			char a[80];
			sprintf(a,"\n%s not found!", file);
            drawString(a,  0, 0, 0xFF0000FF, 0);

			sceGuFinish();
		    sceGuSync(0, 0);

		    sceDisplayWaitVblankStart();
		    sceGuSwapBuffers();
			t++;
		}
		return;
	}
	OGG_Init(1);
	OGG_Load(file);
	memset(ogginfo, 0, 1024);
	OGG_GetInfo(ogginfo);
	OGG_Play();
	was_ogg = 1;

	while(!OGG_EndOfStream() && !(changedButtons & PSP_CTRL_CROSS))
	{
		char a[80],b[80],d[80];

		sceGuStart(GU_DIRECT, list);
		sceGuClear(GU_COLOR_BUFFER_BIT);

        sprintf(b,"Mod Player by Crow_bar for IRSHELL 5.xx\n");
        drawString(b, 0, 16 * 0, 0xFF0000FF, 0);

		sprintf(a,"Playing %s\n", file);
        drawString(a, 0, 16 * 1, 0xFF00FF00, 0);

		drawString("play/pause: start\n",  0, 16 * 2, 0xFF0000FF, 0);
        drawString("stop: cross\n",        0, 16 * 3, 0xFF0000FF, 0);
        drawString("previous/next position: d-pad left/right\n", 0, 16 * 4, 0xFF0000FF, 0);
        drawString("start over: square\n", 0, 16 * 4, 0xFF0000FF, 0);

		memset(ogginfo, 0, 1024);

		OGG_GetTimeString(ogginfo);

		sprintf(d,"%s       seekable: %i", ogginfo, OGG_IsSeekable());
		drawString(d, 0, 16 * 5, 0xFFFFFF00, 0);

		sceCtrlReadBufferPositive(&pad, 1);

		changedButtons = pad.Buttons & (~oldButtons);

		if (changedButtons & PSP_CTRL_START)
			OGG_Pause();

		else if ((changedButtons & PSP_CTRL_LEFT) && OGG_IsSeekable())
			OGG_Rewind();

		else if((changedButtons & PSP_CTRL_RIGHT) && OGG_IsSeekable())
			OGG_FastForward();

		else if(changedButtons & PSP_CTRL_SQUARE)
			OGG_Restart();

		oldButtons = pad.Buttons;

		sceGuFinish();
		sceGuSync(0, 0);

		sceDisplayWaitVblankStart();
		sceGuSwapBuffers();
	}
	OGG_Stop();
	OGG_FreeTune();
}

int main(int argc, char *argv[])
{
	setupCallbacks();
	pspDebugScreenInit();
	pspAudioInit();

	SetupGu();

	MikMod_RegisterAllDrivers();
	MikMod_RegisterAllLoaders();
	MikMod_Init("");

	char *file;
	char ext[6];

	if(argc == 2)
	    file = argv[1];
	else
	{
		pspAudioEnd();
		sceKernelExitGame();
	}
	if(argc == 2)
	{
		if(exists(file))
		{
			get_fileext(file, ext);
			if(!strcmp(ext,"ogg"))
				play_ogg(file);
            if(!strcmp(ext,"mp3") || !strcmp(ext,"MP3"))
				play_mp3(file);
			else if(is_mikmodfile(ext))
				play_mikmod(file);
			else
			{
			  int t = 0;
		      while(t < 100) //~3 sec
		      {
				sceGuStart(GU_DIRECT, list);
		        sceGuClear(GU_COLOR_BUFFER_BIT);

				char a[80];
				sprintf(a,"Unknown format: %s, %s!", file, ext);
                drawString(a,  0, 0, 0xFF0000FF, 0);

				sceGuFinish();
		        sceGuSync(0, 0);

		        sceDisplayWaitVblankStart();
		        sceGuSwapBuffers();
				t++;
			  }
			}
		}
		else
		{
		  int t = 0;
		  while(t < 100) //~3 sec
		  {
			sceGuStart(GU_DIRECT, list);
		    sceGuClear(GU_COLOR_BUFFER_BIT);

			char a[80];
			sprintf(a,"%s not found!", file);
            drawString(a,  0, 0, 0xFF0000FF, 0);

			sceGuFinish();
		    sceGuSync(0, 0);

		    sceDisplayWaitVblankStart();
		    sceGuSwapBuffers();
			t++;
		  }
		}
	}
	else
	{
        int t = 0;
		while(t < 100) //~3 sec
		{
			sceGuStart(GU_DIRECT, list);
		    sceGuClear(GU_COLOR_BUFFER_BIT);

            drawString("No file specified!",  0, 0, 0xFF0000FF, 0);

     		sceGuFinish();
		    sceGuSync(0, 0);

		   sceDisplayWaitVblankStart();
	 	   sceGuSwapBuffers();
		   t++;
	    }
	}
	MikMod_Exit();

	if(was_ogg)
	   OGG_End();

	ShGu();

	pspAudioEnd();
	sceKernelExitGame();

	return 0;
}
