extern "C"
{
#include "../quakedef.h"
}

#include <pspgu.h>
#include <pspgum.h>

#define MAX_DECALS	32


enum
{
	DECAL_NONE,
	DECAL_BULLTMRK,
	DECAL_BLOODMRK1,
	DECAL_BLOODMRK2,
	DECAL_BLOODMRK3,
	DECAL_EXPLOMRK
};


// decals
typedef struct
{
	unsigned char type;
	unsigned int gl_texture;
	double	time;
	float	alpha;
	vec3_t	normal, org;
	vec3_t	v[4];
} decal_t;

int loadtextureimage (char* filename, int matchwidth, int matchheight, qboolean complain, int filter);

decal_t			cl_decals[MAX_DECALS];
extern			client_state_t cl;

static int bulletmarktex = 0;
static int bloodmarktex[3] = {0, 0, 0};
static int explosionmarktex = 0;
int num_decals = 0;

cvar_t gl_polyoffset = {"gl_polyoffset", "-0.1", qfalse};

/*
===============
CL_ClearDecals

===============
*/

void CL_ClearDecals(void)
{
	memset (cl_decals, 0, sizeof(decal_t)*MAX_DECALS);
	num_decals = 0;
}

/*
===============
CL_AllocDecal

Find the oldest decal
===============
*/
decal_t *CL_AllocDecal(int type)
{
	int i, d = -1;
	double besttime = 9999999.0;

	if (type)
	{
		for (i = 0; i < MAX_DECALS; i++)
		{
			// try finding exact match
			if (cl_decals[i].time < besttime)
			{
				if (!cl_decals[i].type || (cl_decals[i].type == type))
				{
					d = i;
					besttime = cl_decals[i].time;
				}
			}
		}

		if (d > -1)
			return &cl_decals[d];
	}

	d = -1;
	besttime = 9999999.0;

	for (i = 0; i < MAX_DECALS; i++)
	{
		if (cl_decals[i].time < besttime) {
			d = i;
			besttime = cl_decals[i].time;
		}
	}

	if (d > -1)
		return &cl_decals[d];

	return &cl_decals[0];
}

/*
==============
CL_AddDecal

==============
*/
static vec3_t axis[3] = 
{
	{ 1.0f, 0.0f, 0.0f },
	{ 0.0f, 1.0f, 0.0f },
	{ 0.0f, 0.0f, 1.0f }
};

void CL_AddDecal (vec3_t n, vec3_t pt, int type)
{
	decal_t *decal;
	vec3_t fn, up, right, t;
	int major = 0;
	int size = 3;

	if ((type < DECAL_NONE) || (type > DECAL_EXPLOMRK))
		return;

	decal = CL_AllocDecal (type);
	num_decals++;

	VectorCopy (n, decal->normal);
	VectorCopy (pt, decal->org);

	fn[0] = fabs(n[0]);
	fn[1] = fabs(n[1]);
	fn[2] = fabs(n[2]);

	// find the major axis
	if (fn[1] > fn[major])
		major = 1;
	if (fn[2] > fn[major])
		major = 2;

	// build right vector by hand
	if (fn[0] == 1 || fn[1] == 1 || fn[2] == 1)
	{
		if ((major == 0 && n[0] > 0) || major == 1)
			VectorSet(right, 0, 0, -1);
		else if (major == 0)
			VectorSet(right, 0, 0, 1);
		else
			VectorSet(right, n[2], 0, 0);
	}
	else {
		CrossProduct (axis[major], n, right);
	}

	CrossProduct (n, right, up);
	VectorNormalize (up);
	VectorNormalize (right);

	VectorSubtract (right, up, t);
	VectorScale (t, size, t);
	VectorAdd (pt, t, decal->v[1]);
	VectorAdd (pt, -t, decal->v[3]);

	VectorAdd (right, up, t);
	VectorScale (t, size, t);
	VectorAdd (pt, t, decal->v[2]);
	VectorAdd (pt, -t, decal->v[0]);

	decal->time = cl.time + 120.0;
	decal->alpha = 1.0f;

	// damn... BUGFIX
	decal->type = type;

	if (type == DECAL_BULLTMRK)
		decal->gl_texture = bulletmarktex;
	else if ((type >= DECAL_BLOODMRK1) && (type <= DECAL_BLOODMRK3))
		decal->gl_texture = bloodmarktex[type - DECAL_BLOODMRK1];
	else if (decal->gl_texture == DECAL_EXPLOMRK)
		decal->gl_texture = explosionmarktex;
}

static int CompareDecals(const void *a, const void *b)
{
    return ((decal_t*)a)->gl_texture - ((decal_t*)b)->gl_texture;
}

static void R_SortDecals(void)
{
    qsort((void *)cl_decals, MAX_DECALS, sizeof(decal_t), CompareDecals);
}

/*
===============
R_RenderDecals

Minimizing GL-calls by sorting decals
===============
*/
void R_RenderDecals (void)
{
	int i, texnum = 0;
	vec3_t v;

	// don't render if there are no decals
	if (!num_decals)
		return;

	// sort decals by texture
	R_SortDecals();

	sceGumPushMatrix();

	sceGuDepthMask (GU_TRUE);
	sceGuEnable(GU_BLEND);
	sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_FIX, 0, 0xFFFFFFFF);
	sceGuTexFunc(GU_TFX_MODULATE , GU_TCC_RGB);


	for (i = 0; i < MAX_DECALS; i++)
	{
		if (cl_decals[i].time > cl.time)
		{
			VectorSubtract(cl_decals[i].org, r_refdef.vieworg, v);
			if (DotProduct(cl_decals[i].normal, v) < 0)
				continue;

			//glColor4f (1.0f, 1.0f, 1.0f, cl_decals[i].alpha);

			   sceGuColor(GU_RGBA(
	             static_cast<unsigned int>(1.0f * 255.0f),
	             static_cast<unsigned int>(1.0f * 255.0f),
	             static_cast<unsigned int>(1.0f * 255.0f),
	             static_cast<unsigned int>(cl_decals[i].alpha * 255.0f)));

				if ((cl_decals[i].gl_texture > 0) &&
				(cl_decals[i].gl_texture != texnum)) {
				texnum = cl_decals[i].gl_texture;
				GL_Bind (texnum);
				//glBindTexture (GL_TEXTURE_2D, texnum);
			}

			// Allocate the vertices.
		    struct vertex
		    {
			float u, v;
			float x, y, z;
		    };

		    vertex* const out = static_cast<vertex*>(sceGuGetMemory(sizeof(vertex) * 4));

			out[0].u = 0.0f;
			out[0].v = 0.0f;
			out[0].x = cl_decals[i].v[0][0];
            out[0].y = cl_decals[i].v[0][1];
            out[0].z = cl_decals[i].v[0][2];

			out[1].u = 1.0f;
			out[1].v = 0.0f;
			out[1].x = cl_decals[i].v[1][0];
            out[1].y = cl_decals[i].v[1][1];
            out[1].z = cl_decals[i].v[1][2];

			out[2].u = 1.0f;
			out[2].v = 1.0f;
			out[2].x = cl_decals[i].v[2][0];
            out[2].y = cl_decals[i].v[2][1];
            out[2].z = cl_decals[i].v[2][2];

			out[3].u = 0.0f;
			out[3].v = 1.0f;
			out[3].x = cl_decals[i].v[3][0];
            out[3].y = cl_decals[i].v[3][1];
            out[3].z = cl_decals[i].v[3][2];
			// Draw the clipped vertices.

		sceGuDrawArray(
		     GU_TRIANGLE_FAN,
		     GU_TEXTURE_32BITF | GU_VERTEX_32BITF,
		     4, 0, out);
		}
		else
		{
			if (cl_decals[i].alpha > 0)
			{
				cl_decals[i].alpha = 0;
				num_decals--;
			}
		}
	}

    sceGuDisable(GU_BLEND);
    sceGuDepthMask(GU_FALSE);
	sceGuTexFunc(GU_TFX_REPLACE, GU_TCC_RGB);
	sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
	sceGumPopMatrix();
    sceGumUpdateMatrix();
}

/*
===============
R_InitDecalsTextures

===============
*/
void R_InitDecalsTextures (void)
{
	//glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	//glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	bulletmarktex = loadtextureimage ("gfx/hole", 0, 0, qtrue, GU_LINEAR);
	explosionmarktex = 0;
}

