//============================================================================

#include "qcc.h"
#include "hash.h"
#include <time.h>

#include <psppower.h>
#include <pspkernel.h>
#include <psprtc.h>
#include <pspdebug.h>
#include <pspdisplay.h>

PSP_MODULE_INFO("DQ_QC_CMD", PSP_MODULE_USER, 1, 0);
PSP_MAIN_THREAD_ATTR(PSP_THREAD_ATTR_USER | PSP_THREAD_ATTR_VFPU);
PSP_HEAP_SIZE_KB(-1024);

#define print pspDebugScreenPrintf

void PrintFunction (char *name);
char commandline[1024];
/* Exit callback */
int exitCallback(int arg1, int arg2, void *common)
{
	sceKernelExitGame();
	return 0;
}

/* Callback thread */
int callbackThread(SceSize args, void *argp)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", (void*) exitCallback, NULL);
	sceKernelRegisterExitCallback(cbid);
	sceKernelSleepThreadCB();

	return 0;
}

/* Sets up the callback thread and returns its thread id */
int setupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", callbackThread, 0x11, 0xFA0, 0, 0);
	if (thid >= 0) {
		sceKernelStartThread(thid, 0, 0);
	}
	return thid;
}

/*
============
main
============
*/

time_t StartTime;
int main (int argc, char **argv)
{
	char	*src;
	char	*src2;
	char	filename[1024];
	int		p, crc, handle;

	int i;

	setupCallbacks();

	scePowerSetClockFrequency(333, 333, 166);

	pspDebugScreenInit();
    pspDebugScreenClear();

	print("---------------PSP frikqcc v2.7 by Crow_bar----------------\n");

	myargc = argc;
	myargv = argv;

	summary = false;
	pr_pause = !(argc > 1);
	strcpy (sourcedir, "");
	logging = true;
	commandline[0]=0;

	for(i = 2; i < argc; i++)
	{
		strcat(commandline, argv[i]);
		strcat(commandline, " ");
	}
	ParseParms(commandline);


	if ( CheckParm ("-?") || CheckParm ("-help") || CheckParm ("-h") || CheckParm("/?"))
	{
		print("frikqcc looks for progs.src in the current directory.\n");
		print("to look in a different directory: frikqcc -src <directory>\n");
		print("to set warning level: frikqcc -warn <warnlevel>\n");
		print("to decompile a progs.dat: frikqcc -dec <datfile>\n");
		print(" \n");
		print("Optimizations:\n");
		print("\t/Ot\tshorten stores\n");
		print("\t/Oi\tshorten ifs\n");
		print("\t/Op\tnon-vector parms\n");
		print("\t/Oc\teliminate constant defs/names\n");
		print("\t/Od\teliminate duplicate defs\n");
		print("\t/Os\thash lookup in CopyString\n");
		print("\t/Ol\teliminate local, static and immediate names\n");
		print("\t/On\teliminate unneeded function names\n");
		print("\t/Of\tstrip filenames\n");
		print("\t/Ou\tremove unreferenced variable defs/names\n");
		print("\t/Oa\tremove constant arithmetic\n");
		print("\t/Oo\tAdd logic jumps\n");
		print("\t/Or\teliminate temps\n");
		print("\t/O1\tuse all optimizations except /Oo\n");
		print("\t/O2\tuse all optimizations\n");
		EndProgram(false);
		return 1;
	}


	remove("error.log");
	InitData ();
	pr_dumpasm = false;

// no cameras past this point
	StartTime = I_FloatTime();

	PR_BeginCompilation (PR_Malloc (0x100000), 0x100000);

	
	sprintf (filename, "%spreprogs.src", sourcedir);
	handle = open(argv[1], O_RDONLY);

	destfile[0] = 0;

	if(handle == -1)
	{
		close(handle);
		sprintf (filename, "%sprogs.src", sourcedir);
		LoadFile (filename, (void **)&src);
	
		src = COM_Parse (src);
		if (!src)
			Sys_Error ("Q606: No destination filename.  frikqcc -help for info.\n");
		strcpy (destfile, com_token);
	}
	else
	{

		close(handle);
		LoadFile (filename, (void **)&src2);

		if (!PR_CompileFile (src2, "preprogs.src") )
		{
			EndProgram(false);
			return 1;
		}
		if(!destfile[0])
		{
			print("Warning: No destination file was set, assuming progs.dat");
			strcpy(destfile, "progs.dat");
		}

		goto endcompilation;
	}



// compile all the files
	do
	{
		src = COM_Parse(src);
		if (!src)
			break;
		sprintf (filename, "%s%s", sourcedir, com_token);
		print ("%s\n", filename);
		LoadFile (filename, (void **)&src2);
		if(!PR_CompileFile (src2, com_token))
			EndProgram(false);
	} while (1);

endcompilation:

	if (!PR_FinishCompilation ())
	{
		EndProgram(false);
		return 1;
	}

	p = CheckParm ("-asm");
	if (p)
	{
		for (p++ ; p<argc ; p++)
		{
			if (argv[p][0] == '-')
				break;
			PrintFunction (argv[p]);
		}
	}



// write progdefs.h

	crc = PR_WriteProgdefs ("progdefs.h");
	if (p = PR_FindDefine("SYSTEM_CRC"))
		crc = pr_defines[p].value._float;

// write data file

	WriteData (crc);
	if (CheckParm("-strings"))
		PrintStrings();
	if (CheckParm("-globals"))
		PrintGlobals();
	if (CheckParm("-functions"))
		PrintFunctions();
	if (CheckParm("-fields"))
		PrintFields();

	//int sum = 0;
	//for (int i = 0 ; i < HSIZE1 ; i++)
	//	sum += stats[i] * stats[i];
	//printf("%6d sum\n", sum);

	EndProgram(true);

	return 1;
}
