// cmdlib.c

/*  Copyright (C) 1996-1997  Id Software, Inc.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    See file, 'COPYING', for details.
*/

#include "qbsp.h"
//#include <direct.h>
#include <sys/types.h>	// gb
#include <sys/timeb.h>	// gb
#include <sys/stat.h>	// gb
#include <string.h>	// gb
#include <unistd.h>	// gb - getwd

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#include "windows.h"
#endif

#include <pspkernel.h>
#include <psprtc.h>

#define printf	pspDebugScreenPrintf

#define PATHSEPERATOR   '/'


char *copystring(char *s)
{
	char	*b;
	b = (char *)AllocMem(OTHER, strlen(s)+1);
	strcpy (b, s);
	return b;
}



/*
================
I_FloatTime
================
*/
double I_FloatTime (void)
{
/*
	struct timeb timebuffer;	// gb, was _timeb

	ftime(&timebuffer);		// gb, was _ftime

	return (double)timebuffer.time + (timebuffer.millitm / 1000.0);
*/
    u64 ticks;
	sceRtcGetCurrentTick(&ticks);
	return ticks * 0.000001;
}

void Q_getwd (char *out)
{
#ifdef WIN32
   _getcwd (out, 256);
   strcat (out, "\\");
#else
   char *pwd;
   int len;

   pwd = getenv("PWD");
   if (!pwd)
   	printf("Error: Couldn't get working directory - "
	       "PWD not set in environment\n");
   len = strlen(pwd);
   if (len > 255)
	   printf("Error: Not enough space to hold working dir\n");
   strcpy(out, pwd);
#endif
}


/*
=============================================================================

						MISC FUNCTIONS

=============================================================================
*/


void DefaultExtension (char *path, char *extension)
{
	char    *src;
//
// if path doesn't have a .EXT, append extension
// (extension should include the .)
//
	src = path + strlen(path) - 1;

	while (*src != PATHSEPERATOR && *src != '\\' && src != path)
	{
		if (*src == '.')
			return;                 // it has an extension
		src--;
	}

	strcat (path, extension);
}


void DefaultPath (char *path, char *basepath)
{
	char    temp[128];

	if (path[0] == PATHSEPERATOR || path[0] == '\\')
		return;                   // absolute path location
	strcpy (temp,path);
	strcpy (path,basepath);
	strcat (path,temp);
}


void    StripFilename (char *path)
{
	int             length;

	length = strlen(path)-1;
	while (length > 0 && path[length] != PATHSEPERATOR && path[length] != '\\')
		length--;
	path[length] = 0;
}

void    StripExtension (char *path)
{
	int             length;

	length = strlen(path)-1;
	while (length > 0 && path[length] != '.')
	{
		length--;
		if (path[length] == PATHSEPERATOR || path[length] == '\\')
			return;		// no extension
	}
	if (length)
		path[length] = 0;
}

void SetQPriority(int Priority)
#ifdef WIN32
{
	if (Priority == 0)
		Priority = THREAD_PRIORITY_BELOW_NORMAL;
	else if (Priority == 1)
		Priority = THREAD_PRIORITY_NORMAL;
	else if (Priority == 2)
		Priority = THREAD_PRIORITY_ABOVE_NORMAL;
	else
		Message(msgError, errInvalidPriority);

	SetThreadPriority(GetCurrentThread(), Priority);
}
#else
{
}
#endif

void ShowBar(int Current, int Total)

{
	double	      Time;
	int	      New;
	static int    Prev = -1;
	static double PrevTime = 0;

	if (Total == 0)
	{
		// Enable
		Prev = 0;
		return;
	}

	if (Prev == -1)
		return; // Disabled

	if (Total == -1)
	{
		if (Current == 0)
		{
			// Disrupted
			if (Prev > 0)
			{
				printf("\\\n");

				// Make sure bar will be repainted next time
				Prev = 0;
				PrevTime = -1;
			}

			return;
		}

		// Finish and disable

		if (options.fNumpercent)
			printf("\r%3d%%\n", 100);
		else
		{
			// Finish bar
			while (Prev < 9)
				printf("%c", ++Prev == 5 ? '+' : '-');

			printf("+\n");
    		}

		fflush(stdout);

		Prev = -1;
		PrevTime = 0;

		return;
	}

	if (options.fNumpercent)
	{
		// Update every 20th percent or once each second
		New = Current * 100 / Total;

		if (New <= Prev || New > 99)
			return;

		Time = I_FloatTime();

		if (New / 20 == Prev / 20 && Time < PrevTime + 1)
			return;

		Prev = New;
		PrevTime = Time;
		printf("\r%3d%%", New);
	}
	else
	{
		// Update every 10th percent
		New = Current * 10 / Total;

		if (New <= Prev || New > 9)
			return;

		// Make sure bar reaches current value
		do
			printf("%c", ++Prev == 5 ? '+' : '-');
		while (Prev < New);
	}

	fflush(stdout);
}
