/*  Copyright (C) 1996-1997  Id Software, Inc.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    See file, 'COPYING', for details.
*/


#include "../include/common/cmdlib.h"
#include "../include/common/bspfile.h"

#include <pspkernel.h>
#include <psppower.h>
#include <pspdebug.h>
#include <pspdisplay.h>
#include <pspctrl.h>

PSP_MODULE_INFO("DQ_BSPINFO", PSP_MODULE_USER, 1, 0);
PSP_MAIN_THREAD_ATTR(PSP_THREAD_ATTR_USER | PSP_THREAD_ATTR_VFPU);
PSP_HEAP_SIZE_KB(-1024);

#define printf	pspDebugScreenPrintf

/* Exit callback */
int exitCallback(int arg1, int arg2, void *common)
{
	sceKernelExitGame();
	return 0;
}

/* Callback thread */
int callbackThread(SceSize args, void *argp)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", (void*) exitCallback, NULL);
	sceKernelRegisterExitCallback(cbid);
	sceKernelSleepThreadCB();

	return 0;
}

/* Sets up the callback thread and returns its thread id */
int setupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", callbackThread, 0x11, 0xFA0, 0, 0);
	if (thid >= 0) {
		sceKernelStartThread(thid, 0, 0);
	}
	return thid;
}

int main(int argc, char **argv)
{
    int i;
    char source[1024];

	setupCallbacks();

	scePowerSetClockFrequency(333, 333, 166);

	pspDebugScreenInit();
    pspDebugScreenClear();

   	//setup Pad
	SceCtrlData pad;
	u32 oldButtons = 0;
	sceCtrlSetSamplingCycle(0);
	sceCtrlSetSamplingMode(0);

	if (argc == 1)
	   Error("usage: bspinfo bspfile [bspfiles]");

    for (i = 1; i < argc; i++)
	{
	  printf("---------------------\n");
	  strcpy(source, argv[i]);
	  DefaultExtension(source, ".bsp");
	  printf("%s\n", source);

	  LoadBSPFile(source);
	  PrintBSPFileSizes();
	  printf("---------------------\n");
    }

    printf("Press CROSS to exit\n");

	int done = 0;
	while(!done)
	{
       sceCtrlReadBufferPositive(&pad, 1);
		if(oldButtons != pad.Buttons)
		{
			if(pad.Buttons & PSP_CTRL_CROSS)
				done = 1;
		}
		oldButtons = pad.Buttons;
	}

    sceKernelExitGame();

	return 0;
}
