/*
Copyright (C) 2007 Peter Mackay and Chris Swindle.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
/*
#include <stdio.h>
#include <mikmod.h>
#include <pspaudiolib.h>
#include <stdio.h>
#include <pspmp3.h>
#include <psputility.h>
*/
#include <algorithm>
#include <cstddef>

#include <pspkernel.h>
#include <psputility.h>
#include <pspdebug.h>
#include <pspctrl.h>
#include <pspdisplay.h>
#include <stdio.h>
#include <pspaudio.h>
#include <pspmp3.h>
#include <pspdebug.h>

#include "mp3.h"

extern "C"
{
#include "../quakedef.h"
}

extern 	cvar_t bgmtype;
extern	cvar_t bgmvolume;


namespace quake
{
	namespace cd
	{
		struct Sample
		{
			short left;
			short right;
		};

		static int fd;
		static int handle;
        static int channel = -1;
	    static int samplingRate;
	    static int numChannels;
	    static int lastDecoded = 0;
	    static int mvolume = PSP_AUDIO_VOLUME_MAX;
	    static int numPlayed = 0;
	    static int paused = 0;
	    static int lastButtons = 0;
	    static int loop = 0;


		static int		     	last_track = 1;
        static byte 	remap[100];
        static int	    playTrack = 1;
		static bool	 playing  = false;
		//static bool	 paused   = false;
		static bool	 enabled  = false;
		static float cdvolume = 0;


	}
}

#define MP3FILE "ms0:/MUSIC/Test.mp3"

using namespace quake;
using namespace quake::cd;

// Input and Output buffers
char	mp3Buf[16*1024]  __attribute__((aligned(64)));
short	pcmBuf[16*(1152/2)]  __attribute__((aligned(64)));

int fillStreamBuffer( int fd, int handle )
{
	char* dst;
	int write;
	int pos;
	// Get Info on the stream (where to fill to, how much to fill, where to fill from)
	int status = sceMp3GetInfoToAddStreamData( handle, (SceUChar8**)(&dst), &write, &pos);
	if (status<0)
	{
		Con_Printf("sceMp3GetInfoToAddStreamData returned 0x%08X\n", status);
	}

	// Seek file to position requested
	status = sceIoLseek32( fd, pos, SEEK_SET );
	if (status<0)
	{
		Con_Printf("sceIoLseek32 returned 0x%08X\n", status);
	}

	// Read the amount of data
	int read = sceIoRead( fd, dst, write );
	if (read < 0)
	{
		Con_Printf("Could not read from file - 0x%08X\n", read);
	}

	if (read==0)
	{
		// End of file?
		return 0;
	}

	// Notify mp3 library about how much we really wrote to the stream buffer
	status = sceMp3NotifyAddStreamData( handle, read );
	if (status<0)
	{
		Con_Printf("sceMp3NotifyAddStreamData returned 0x%08X\n", status);
	}

	return (pos>0);
}

static void CD_f (void)
{
	char	*command;
    int		ret;
	int		n;

	if (Cmd_Argc() < 2)
	{
		Con_Printf("commands:");
		Con_Printf("on, off, reset, remap, \n");
		Con_Printf("play, stop, loop, pause, resume\n");
		Con_Printf("eject, close, info\n");
		return;
	}

	command = Cmd_Argv (1);

	if (Q_strcasecmp(command, "on") == 0)
	{
		enabled = true;
		return;
	}

	if (Q_strcasecmp(command, "off") == 0)
	{
		if (playing)
			CDAudio_Stop();
		enabled = false;
		return;
	}

	if (Q_strcasecmp(command, "reset") == 0)
	{
		enabled = true;
		if (playing)
			CDAudio_Stop();
        for (n = 0; n < 100; n++)
			remap[n] = n;
		return;
	}

	if (Q_strcasecmp(command, "remap") == 0)
	{
        ret = Cmd_Argc() - 2;
		if (ret <= 0)
		{
			for (n = 1; n < 100; n++)
				if (remap[n] != n)
					Con_Printf("  %u -> %u\n", n, remap[n]);
			return;
		}
		for (n = 1; n <= ret; n++)
			remap[n] = Q_atoi(Cmd_Argv (n+1));
		return;
	}

	if (Q_strcasecmp(command, "close") == 0)
	{
		return;
	}

	if (Q_strcasecmp(command, "play") == 0)
	{
		CDAudio_Play((byte)atoi(Cmd_Argv (2)), (qboolean) false);
		return;
	}

	if (Q_strcasecmp(command, "loop") == 0)
	{
		CDAudio_Play((byte)atoi(Cmd_Argv (2)), (qboolean) true);
		return;
	}

	if (Q_strcasecmp(command, "stop") == 0)
	{
		CDAudio_Stop();
		return;
	}

	if (Q_strcasecmp(command, "pause") == 0)
	{
		CDAudio_Pause();
		return;
	}

	if (Q_strcasecmp(command, "resume") == 0)
	{
		CDAudio_Resume();
		return;
	}

	if (Q_strcasecmp(command, "eject") == 0)
	{
		if (playing)
			CDAudio_Stop();
		return;
	}

	if (Q_strcasecmp(command, "info") == 0)
	{
		return;
	}

}

void CDAudio_VolumeChange(float bgmvolume)
{
	//int volume = (int) (bgmvolume * (float) PSP_VOLUME_MAX);
    //changeMp3Volume=0;
	//cdvolume = bgmvolume;
}

void CDAudio_Play(byte track, qboolean looping)
{
    if (!enabled)
		return;
    // Check if we need to fill our stream buffer
	if (sceMp3CheckStreamDataNeeded( handle )>0)
	{
		fillStreamBuffer( fd, handle );
	}

	// Decode some samples
	short* buf;
	int bytesDecoded;
	int retries = 0;
	// We retry in case it's just that we reached the end of the stream and need to loop
	for (;retries<1;retries++)
	{
		bytesDecoded = sceMp3Decode( handle, &buf );
		if (bytesDecoded>0)
			break;

		if (sceMp3CheckStreamDataNeeded( handle )<=0)
			break;

		if (!fillStreamBuffer( fd, handle ))
		{
			numPlayed = 0;
		}
	}
	if (bytesDecoded<0 && bytesDecoded!=0x80671402)
	{
		Con_Printf("ERROR: sceMp3Decode returned 0x%08X\n", bytesDecoded);
	}

	// Nothing more to decode? Must have reached end of input buffer
	if (bytesDecoded==0 || bytesDecoded==0x80671402)
    {
		paused = 1;
		sceMp3ResetPlayPosition( handle );
		numPlayed = 0;
	}
	else
	{
	    // Reserve the Audio channel for our output if not yet done
	    if (channel<0 || lastDecoded!=bytesDecoded)
	    {
		  if (channel>=0)
		    sceAudioSRCChRelease();

		  channel = sceAudioSRCChReserve( bytesDecoded/(2*numChannels), samplingRate, numChannels );
	    }
		// Output the decoded samples and accumulate the number of played samples to get the playtime
		numPlayed += sceAudioSRCOutputBlocking( mvolume, buf );
	}
	
}

void CDAudio_Stop(void)
{
    if (!playing)
		return;

	if (!enabled)
		return;
    
	playing = false;
}

void CDAudio_Pause(void)
{
	
	paused = true;
}

void CDAudio_Resume(void)
{
	
	paused = false;
}

void CDAudio_Update(void)
{
	// Fill the input buffer.
	if (strcmpi(bgmtype.string,"cd") == 0) {
		if (playing == false) {
			CDAudio_Play(playTrack, (qboolean) false);
		}
		if (paused == true) {
			CDAudio_Resume();
		}
	
	} else {
		if (paused == false) {
			CDAudio_Pause();
		}
		if (playing == true) {
			CDAudio_Stop();
		}
	}
}

int CDAudio_Init(void)
{
	if (cls.state == ca_dedicated)
		return -1;

	if (COM_CheckParm("-nocdaudio"))
		return -1;

	// Load modules
	int status = sceUtilityLoadModule(PSP_MODULE_AV_AVCODEC);
	if (status<0)
	{
		Con_Printf("sceUtilityLoadModule(PSP_MODULE_AV_AVCODEC) returned 0x%08X\n", status);
	}

	status = sceUtilityLoadModule(PSP_MODULE_AV_MP3);
	if (status<0)
	{
		Con_Printf("sceUtilityLoadModule(PSP_MODULE_AV_MP3) returned 0x%08X\n", status);
	}

	// Open the input file
	fd = sceIoOpen(MP3FILE, PSP_O_RDONLY, 0777 );
	if (fd<0)
	{
		Con_Printf("Could not open file '%s' - 0x%08X\n", MP3FILE, fd);
	}

	// Init mp3 resources
	status = sceMp3InitResource();
	if (status<0)
	{
		Con_Printf("sceMp3InitResource returned 0x%08X\n", status);
	}

	// Reserve a mp3 handle for our playback
	SceMp3InitArg mp3Init;
	mp3Init.mp3StreamStart = 0;
	mp3Init.mp3StreamEnd = sceIoLseek32( fd, 0, SEEK_END );
	mp3Init.unk1 = 0;
	mp3Init.unk2 = 0;
	mp3Init.mp3Buf = mp3Buf;
	mp3Init.mp3BufSize = sizeof(mp3Buf);
	mp3Init.pcmBuf = pcmBuf;
	mp3Init.pcmBufSize = sizeof(pcmBuf);

	handle = sceMp3ReserveMp3Handle( &mp3Init );
	if (handle<0)
	{
		Con_Printf("sceMp3ReserveMp3Handle returned 0x%08X\n", handle);
	}

	// Fill the stream buffer with some data so that sceMp3Init has something to work with
	fillStreamBuffer( fd, handle );

	status = sceMp3Init( handle );
	if (status<0)
	{
		Con_Printf("sceMp3Init returned 0x%08X\n", status);
	}

	samplingRate = sceMp3GetSamplingRate( handle );
	numChannels = sceMp3GetMp3ChannelNum( handle );
	
	enabled = true;

	Cmd_AddCommand ("cd", CD_f);
	
	Con_Printf("CD Audio Initialized\n");

	return 0;
}

void CDAudio_Shutdown(void)
{
	Con_Printf("CDAudio_Shutdown\n");
	
	CDAudio_Stop();

}

